// ReadLine.pp
// Written by Bill Stewart (bstewart@iname.com)
//
// Short Free Pascal (http://www.freepascal.org/) console application that
// reads console input and writes it to standard output.
//
// Input is forced to come from the console (ignoring redirection), but output
// is written to standard output (allowing redirection).
//
// You can hide input, specify a prompt, and (if not hiding input) specify
// an initial input value.

{$MODE OBJFPC}
{$H+}

{$APPTYPE CONSOLE}
{$R *.RC}

Uses
  GetOpts, Strings, Windows;

Const
  MAX_INPUT_LENGTH = 8190;

Type
  // ReadLn includes CR+LF at end, so +1; hence max number of characters
  // allowed in input buffer is MAX_INPUT_LENGTH
  INPUT_BUFFER = Array[0..MAX_INPUT_LENGTH + 1] Of Char;

Var
  Hidden: Boolean;                    // -h
  Prompt: String;                     // -p
  Opt: Char;                          // GetOpts option character
  ConsoleOutput, ConsoleInput: Text;  // fpc files
  hConsoleInput: HANDLE;              // win32 console handle
  ConsoleMode: DWORD;                 // win32 console mode
  InitialValue: String;               // initial input value
  InputString: String;                // input string
  InputBuffer: INPUT_BUFFER;          // input buffer

// Returns pointer to command line, starting at specified argument
Function GetCommandTail(Const StartArg: DWORD): PChar;
  Const
    WHITESPACE: Set Of Char = [#9, #32];
  Var
    pCL, pTail: PChar;
    InQuote: Boolean;
    ArgNo, N: DWORD;
  Begin
  pCL := GetCommandLine;
  pTail := Nil;
  If pCL^ <> #0 Then
    Begin
    While pCL^ In WHITESPACE Do  //Skip leading whitespace
      Inc(pCL);
    InQuote := False;
    pTail := pCL;
    ArgNo := 0;
    For N := 0 To StrLen(pCL) Do
      Case pCL[N] Of
        #0:
          Break;
        '"':
          Begin
          InQuote := Not InQuote;
          If InQuote Then
            Begin
            If ArgNo = StartArg Then
              Break;
            End;
          Inc(pTail);
          End;
        #9, #32:
          Begin
          If (Not InQuote) And (Not (pCL[N - 1] In WHITESPACE)) Then
            Inc(ArgNo);
          Inc(pTail);
          End;
        Else
          Begin
          If ArgNo = StartArg Then
            Break
          Else
            Inc(pTail);
          End;
      End;
    End;
  Result := pTail;
  End;

// Pre-populates the specified console input handle's buffer with a string
Procedure PopulateConsoleInputBuffer(hConsoleInput: HANDLE; S: String);
  Var
    N, NumWritten: DWORD;
    InputRecord: INPUT_RECORD;
  Begin
  For N := 1 To Length(S) Do
    Begin
    FillChar(InputRecord, SizeOf(InputRecord), 0);
    InputRecord.EventType := KEY_EVENT;
    InputRecord.Event.KeyEvent.bKeyDown := True;
    InputRecord.Event.KeyEvent.wRepeatCount := 1;
    InputRecord.Event.Keyevent.AsciiChar := S[N];
    WriteConsoleInput(hConsoleInput,  // HANDLE hConsoleInput
                      InputRecord,    // INPUT_RECORD lpBuffer
                      1,              // nLength
                      NumWritten);    // lpNumberOfEventsWritten
    InputRecord.Event.KeyEvent.bKeyDown := False;
    WriteConsoleInput(hConsoleInput,  // HANDLE hConsoleInput
                      InputRecord,    // INPUT_RECORD lpBuffer
                      1,              // nLength
                      NumWritten);    // lpNumberOfEventsWritten
    End;
  End;

Begin
  if ParamStr(1) = '/?' Then
    Begin
    WriteLn('ReadLine - (C) 2014 by Bill Stewart (bstewart@iname.com)');
    WriteLn;
    WriteLn('Reads a string from console input and writes it to standard output.');
    WriteLn;
    WriteLn('Usage: ReadLine [-h] [-p prompt] [-- initial_value]');
    WriteLn;
    WriteLn('-h  Hidden input. Characters are not displayed as they are typed.');
    WriteLn('-p  Display a prompt. Enclose it in quotes (") if it contains spaces.');
    WriteLn;
    WriteLn('You can specify an initial input value at the end of the command line after two');
    WriteLn('dashes (--). The initial input value is ignored if you specify -h.');
    WriteLn;
    WriteLn('The prompt can be up to 255 characters long, and the input value can be up to');
    WriteLn(MAX_INPUT_LENGTH, ' characters long.');
    WriteLn;
    WriteLn('The initial input value is used without any parsing or interpretation, so it');
    WriteLn('can include spaces and quotes. The only limitation is that leading whitespace');
    WriteLn('is ignored.');
    Exit;
    End;

  // Set defaults
  Hidden := False;
  Prompt := '';

  OptErr := False;  // don't complain about invalid options
  Repeat
    Opt := GetOpt('HhP:p:');
    Case Opt Of
      'H', 'h':
        Hidden := True;
      'P', 'p':
        Begin
        Prompt := OptArg;
        CharToOemBuff(PChar(Prompt), PChar(Prompt), Length(Prompt));
        End;
    End; //Case
  Until Opt = EndOfOptions;

  // Open console output for writing (ignore redirection)
  Assign(ConsoleOutput, 'CONOUT$');
  Rewrite(ConsoleOutput);

  // Write the prompt
  If Length(Prompt) > 0 Then
    Begin
    Write(ConsoleOutput, Prompt);
    Flush(ConsoleOutput);
    End;

  // Open console input handle
  hConsoleInput := CreateFile('CONIN$',                       // LPCTSTR lpFileName
                              GENERIC_READ Or GENERIC_WRITE,  // DWORD dwDesiredAccess
                              FILE_SHARE_READ,                // DWORD dwShareMode
                              Nil,                            // LPSECURITY_ATTRIBUTES lpSecurityAttributes
                              OPEN_EXISTING,                  // DWORD dwCreationDisposition
                              0,                              // DWORD dwFlagsAndAttributes
                              0);                             // HANDLE hTemplateFile

  // Hidden = disable input echo
  If Hidden Then
    Begin
    GetConsoleMode(hConsoleInput, ConsoleMode);
    SetConsoleMode(hConsoleInput, ConsoleMode And (Not ENABLE_ECHO_INPUT));
    End;

  // Open console input for reading and set input buffer
  Assign(ConsoleInput, 'CONIN$');
  Reset(ConsoleInput);
  SetTextBuf(ConsoleInput, InputBuffer);

  If Not Hidden Then
    Begin
    InitialValue := GetCommandTail(OptInd);
    If Length(InitialValue) > 0 Then
      Begin
      // Pre-populate input buffer with remainder of command line
      CharToOemBuff(PChar(InitialValue), PChar(InitialValue), Length(InitialValue));
      PopulateConsoleInputBuffer(hConsoleInput, InitialValue);
      End;
    End;

  ReadLn(ConsoleInput, InputString);

  If (Length(Prompt) > 0) And Hidden Then
    Begin
    WriteLn(ConsoleOutput);
    Flush(ConsoleOutput);
    End;

  WriteLn(InputString);

  // Overwrite input string
  If Hidden Then
    FillChar(InputString[1], Length(InputString), #0);

End.
